/*
 * JBoss, Home of Professional Open Source
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.cache.config.parsing;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.config.BuddyReplicationConfig;
import org.jboss.cache.config.CacheLoaderConfig;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.ConfigurationException;
import org.jboss.cache.config.CustomInterceptorConfig;
import org.jboss.cache.config.OldFileFormatException;
import org.jboss.cache.config.parsing.element.BuddyElementParser;
import org.jboss.cache.config.parsing.element.CustomInterceptorsElementParser;
import org.jboss.cache.config.parsing.element.EvictionElementParser;
import org.jboss.cache.config.parsing.element.LoadersElementParser;
import org.jboss.cache.lock.IsolationLevel;
import org.jboss.cache.util.FileLookup;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import java.io.InputStream;
import java.util.List;

/**
 * Reads in XMLconfiguration files and spits out a {@link org.jboss.cache.config.Configuration} object.
 * By default this class uses a validating parser (configurable).
 * <p/>
 * Following system properties can be used for customizing parser behavior:
 * <ul>
 *   <li> <b>-Djbosscache.config.validate=false</b> will make the parser non-validating </li>
 *   <li> <b>-Djbosscache.config.schemaLocation=url</b> allows one to specify a validation schema that would override the one specified in the the xml document </li>
 * </ul>
 * This class is stateful and one instance should be used for parsing a single configuration file.
 *
 * @author Mircea.Markus@jboss.com
 * @since 3.0
 */
public class XmlConfigurationParser extends XmlParserBase
{
   private static final Log log = LogFactory.getLog(XmlConfigurationParser.class);

   public static final String VALIDATING_SYSTEM_PROPERTY = "jbosscache.config.validate";
   public static final String SCHEMA_LOCATION_SYSTEM_PROPERTY = "jbosscache.config.schemaLocation";

   /**
    * the resulting configuration.
    */
   private Configuration config = new Configuration();
   private Element root;
   private ErrorHandler errorHandler;
   private boolean isValidating;
   private String schemaLocation;

   /**
    * If validation is on (default) one can specify an error handler for handling validation errors.
    * The default error handler just logs parsing errors received.
    */
   public XmlConfigurationParser(ErrorHandler errorHandler)
   {
      this.errorHandler = errorHandler;
      isValidating = System.getProperty(VALIDATING_SYSTEM_PROPERTY) == null || Boolean.getBoolean(VALIDATING_SYSTEM_PROPERTY);
      schemaLocation = System.getProperty(SCHEMA_LOCATION_SYSTEM_PROPERTY);
   }

   /**
    * Same as {@link #XmlConfigurationParser(org.xml.sax.ErrorHandler)}.
    *
    * @param validating should the underlaying parser disable the validation?
    */
   public XmlConfigurationParser(boolean validating, ErrorHandler errorHandler)
   {
      isValidating = validating;
      this.errorHandler = errorHandler;
   }

   /**
    * Constructs a parser having validation enabled with a ErrorHandler that only logs the parser errors.
    */
   public XmlConfigurationParser()
   {
      this(new LoggingErrorHandler());
   }

   /**
    * Parses an XML file and returns a new configuration.
    * For looking up the file, {@link org.jboss.cache.util.FileLookup} is used.
    *
    * @throws OldFileFormatException if the file has a 2.x format
    * @see org.jboss.cache.util.FileLookup
    */
   public Configuration parseFile(String filename)
   {
      InputStream is = new FileLookup().lookupFile(filename);
      if (is == null)
      {
         throw new ConfigurationException("Unable to find config file " + filename + " either in classpath or on the filesystem!");
      }
      return parseStream(is);
   }

   /**
    * Similar to {@link #parseFile(String)}, just that it does not create the input stream.
    *
    * @throws OldFileFormatException if the file has a 2.x format
    */
   public Configuration parseStream(InputStream configStream)
   {
      readRoot(configStream);
      return processElements();
   }

   /**
    * Root should be the <b>jbosscache</b> element in the configuration file.
    *
    * @throws OldFileFormatException if the file has a 2.x format
    */
   public Configuration parseElement(Element root)
   {
      this.root = root;
      this.root.normalize();
      return processElements();
   }

   public boolean isValidating()
   {
      return isValidating;
   }

   private Configuration processElements()
   {
      if ("server".equalsIgnoreCase(root.getNodeName()))
      {
         throw new OldFileFormatException();
      }
      try
      {
         configureLocking(getSingleElement("locking"));
         configureTransaction(getSingleElement("transaction"));
         configureReplication(getSingleElement("replication"));
         configureSerialization(getSingleElement("serialization"));
         configureInvalidation(getSingleElement("invalidation"));
         configureStartup(getSingleElement("startup"));
         configureStateRetrieval(getSingleElement("stateRetrieval"));
         configureTransport(getSingleElement("transport"));
         configureShutdown(getSingleElement("shutdown"));
         configureJmxStatistics(getSingleElement("jmxStatistics"));
         configureEviction(getSingleElement("eviction"));
         configureCacheLoaders(getSingleElement("loaders"));
         configureCustomInterceptors(getSingleElement("customInterceptors"));
         configureListeners(getSingleElement("listeners"));
      }
      catch (Exception e)
      {
         throw new ConfigurationException("Unexpected excetion while parsing the configuration file", e);
      }
      return config;
   }

   private void configureStateRetrieval(Element element)
   {
      if (element == null) return; //we might not have this configured
      String fetchInMemoryState = getAttributeValue(element, "fetchInMemoryState");
      if (existsAttribute(fetchInMemoryState)) config.setFetchInMemoryState(getBoolean(fetchInMemoryState));
      String stateRetrievalTimeout = getAttributeValue(element, "timeout");
      if (existsAttribute(stateRetrievalTimeout)) config.setStateRetrievalTimeout(getLong(stateRetrievalTimeout));

   }

   private void configureTransaction(Element element)
   {
      if (element == null) return;
      String attrName = "transactionManagerLookupClass";
      String txMngLookupClass = getAttributeValue(element, attrName);
      if (existsAttribute(txMngLookupClass)) config.setTransactionManagerLookupClass(txMngLookupClass);
      String syncRollbackPhase = getAttributeValue(element, "syncRollbackPhase");
      if (existsAttribute(syncRollbackPhase)) config.setSyncRollbackPhase(getBoolean(syncRollbackPhase));
      String syncCommitPhase = getAttributeValue(element, "syncCommitPhase");
      if (existsAttribute(syncCommitPhase)) config.setSyncCommitPhase(getBoolean(syncCommitPhase));
   }

   private void configureSerialization(Element element)
   {
      if (element == null) return;
      String objectInputStreamPoolSize = getAttributeValue(element, "objectInputStreamPoolSize");
      if (existsAttribute(objectInputStreamPoolSize))
         config.setObjectInputStreamPoolSize(getInt(objectInputStreamPoolSize));
      String objectOutputStreamPoolSize = getAttributeValue(element, "objectOutputStreamPoolSize");
      if (existsAttribute(objectOutputStreamPoolSize))
         config.setObjectOutputStreamPoolSize(getInt(objectOutputStreamPoolSize));
      String version = getAttributeValue(element, "version");
      if (existsAttribute(version)) config.setReplVersionString(version);
      String marshallerClass = getAttributeValue(element, "marshallerClass");
      if (existsAttribute(marshallerClass)) config.setMarshallerClass(marshallerClass);
      String useLazyDeserialization = getAttributeValue(element, "useLazyDeserialization");
      if (existsAttribute(useLazyDeserialization)) config.setUseLazyDeserialization(getBoolean(useLazyDeserialization));
      String useRegionBasedMarshalling = getAttributeValue(element, "useRegionBasedMarshalling");
      if (existsAttribute(useRegionBasedMarshalling))
         config.setUseRegionBasedMarshalling(getBoolean(useRegionBasedMarshalling));
   }

   private void configureCustomInterceptors(Element element)
   {
      if (element == null) return; //this element might be missing
      CustomInterceptorsElementParser parser = new CustomInterceptorsElementParser();
      List<CustomInterceptorConfig> interceptorConfigList = parser.parseCustomInterceptors(element);
      config.setCustomInterceptors(interceptorConfigList);
   }

   private void configureListeners(Element element)
   {
      if (element == null) return; //this element is optional
      String asyncPoolSizeStr = getAttributeValue(element, "asyncPoolSize");
      if (asyncPoolSizeStr != null && !asyncPoolSizeStr.trim().equals(""))
      {
         try
         {
            config.setListenerAsyncPoolSize(Integer.parseInt(asyncPoolSizeStr));
         }
         catch (NumberFormatException nfe)
         {
            throw new ConfigurationException("Unable to parse the asyncPoolSize attribute of the listeners element.  Was [" + asyncPoolSizeStr + "]");
         }
      }
   }

   private void configureBuddyReplication(Element element)
   {
      if (element == null) return;//buddy config might not exist, expect that
      BuddyElementParser buddyElementParser = new BuddyElementParser();
      BuddyReplicationConfig brConfig = buddyElementParser.parseBuddyElement(element);
      config.setBuddyReplicationConfig(brConfig);
   }

   private void configureCacheLoaders(Element element)
   {
      if (element == null) return; //null cache loaders are allowed
      LoadersElementParser clElementParser = new LoadersElementParser();
      CacheLoaderConfig cacheLoaderConfig = clElementParser.parseLoadersElement(element);
      config.setCacheLoaderConfig(cacheLoaderConfig);
   }

   private void configureEviction(Element element)
   {
      if (element == null) return; //no eviction might be configured
      EvictionElementParser evictionElementParser = new EvictionElementParser();
      config.setEvictionConfig(evictionElementParser.parseEvictionElement(element));
   }

   private void configureJmxStatistics(Element element)
   {
      if (element == null) return; //might not be specified
      String enabled = getAttributeValue(element, "enabled");
      if (enabled != null) config.setExposeManagementStatistics(getBoolean(enabled));
   }

   private void configureShutdown(Element element)
   {
      if (element == null) return;
      String hookBehavior = getAttributeValue(element, "hookBehavior");
      if (hookBehavior != null) config.setShutdownHookBehavior(hookBehavior);
   }

   private void configureTransport(Element element)
   {
      if (element == null) return; //transport might be missing
      String clusterName = getAttributeValue(element, "clusterName");
      config.setClusterName(clusterName);
      String multiplexerStack = getAttributeValue(element, "multiplexerStack");
      if (existsAttribute(multiplexerStack)) config.setMultiplexerStack(multiplexerStack);
      Element clusterConfig = getSingleElement("jgroupsConfig", element);
      if (clusterConfig != null)
      {
         JGroupsStackParser stackParser = new JGroupsStackParser();
         String clusterConfigStr = stackParser.parseClusterConfigXml(clusterConfig);
         if (clusterConfigStr != null && clusterConfigStr.trim().length() > 0)
            config.setClusterConfig(clusterConfigStr);
      }
   }

   private void configureStartup(Element element)
   {
      if (element == null) return; //we might not have this configured
      String inactiveOnStartup = getAttributeValue(element, "regionsInactiveOnStartup");
      if (existsAttribute(inactiveOnStartup)) config.setInactiveOnStartup(getBoolean(inactiveOnStartup));
   }

   private void configureInvalidation(Element element)
   {
      if (element == null) return; //might be replication
      Element async = getSingleElement("async");
      if (async != null)
      {
         config.setCacheMode(Configuration.CacheMode.INVALIDATION_ASYNC);
         configureAsyncMode(getSingleElement("async", element));
      }
      Element sync = getSingleElement("sync");
      if (sync != null)
      {
         config.setCacheMode(Configuration.CacheMode.INVALIDATION_SYNC);
         configureSyncMode(getSingleElement("sync", element));
      }
   }

   private void configureReplication(Element element)
   {
      if (element == null) return; //might be we do not have a replication enabled
      Element async = getSingleElement("async");
      if (async != null)
      {
         config.setCacheMode(Configuration.CacheMode.REPL_ASYNC);
         configureAsyncMode(getSingleElement("async", element));
      }
      Element sync = getSingleElement("sync");
      if (sync != null)
      {
         config.setCacheMode(Configuration.CacheMode.REPL_SYNC);
         configureSyncMode(getSingleElement("sync", element));
      }
      configureBuddyReplication(getSingleElement("buddy", element));
   }

   private void configureSyncMode(Element element)
   {
      String replTimeout = getAttributeValue(element, "replTimeout");
      if (replTimeout != null) config.setSyncReplTimeout(getLong(replTimeout));
   }

   private void configureAsyncMode(Element element)
   {
      String useReplQueue = getAttributeValue(element, "useReplQueue");
      if (existsAttribute(useReplQueue)) config.setUseReplQueue(getBoolean(useReplQueue));
      String replQueueInterval = getAttributeValue(element, "replQueueInterval");
      if (existsAttribute(replQueueInterval)) config.setReplQueueInterval(getLong(replQueueInterval));
      String replQueueMaxElements = getAttributeValue(element, "replQueueMaxElements");
      if (existsAttribute(replQueueMaxElements)) config.setReplQueueMaxElements(getInt(replQueueMaxElements));
   }

   private void configureLocking(Element element)
   {
      String isolationLevel = getAttributeValue(element, "isolationLevel");
      if (existsAttribute(isolationLevel)) config.setIsolationLevel(IsolationLevel.valueOf(isolationLevel));
      String lockParentForChildInsertRemove = getAttributeValue(element, "lockParentForChildInsertRemove");
      if (existsAttribute(lockParentForChildInsertRemove))
         config.setLockParentForChildInsertRemove(getBoolean(lockParentForChildInsertRemove));
      String lockAcquisitionTimeout = getAttributeValue(element, "lockAcquisitionTimeout");
      if (existsAttribute(lockAcquisitionTimeout)) config.setLockAcquisitionTimeout(getLong(lockAcquisitionTimeout));
      String nodeLockingScheme = getAttributeValue(element, "nodeLockingScheme");
      if (existsAttribute(nodeLockingScheme)) config.setNodeLockingScheme(nodeLockingScheme);
      String writeSkewCheck = getAttributeValue(element, "writeSkewCheck");
      if (existsAttribute(writeSkewCheck)) config.setWriteSkewCheck(getBoolean(writeSkewCheck));
      String concurrencyLevel = getAttributeValue(element, "concurrencyLevel");
      if (existsAttribute(concurrencyLevel)) config.setConcurrencyLevel(getInt(concurrencyLevel));
   }

   private Element getSingleElement(String elementName)
   {
      return getSingleElement(elementName, root);
   }

   private void readRoot(InputStream config)
   {
      try
      {
         DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory.newInstance();
         if (isValidating)
         {
            docBuilderFactory.setValidating(true);
            docBuilderFactory.setNamespaceAware(true);
            docBuilderFactory.setAttribute("http://java.sun.com/xml/jaxp/properties/schemaLanguage", "http://www.w3.org/2001/XMLSchema");
            if (schemaLocation != null)
            {
               if (log.isTraceEnabled()) log.trace("Using the schema location set to: '" + schemaLocation + '\'');
               docBuilderFactory.setAttribute("http://java.sun.com/xml/jaxp/properties/schemaSource", schemaLocation);
            }
            else if (log.isTraceEnabled())
            {
               log.trace("Validation is enabled, using the schema decalred in the .xml configuration file");
            }
         }
         DocumentBuilder parser = docBuilderFactory.newDocumentBuilder();
         parser.setErrorHandler(errorHandler);
         Document doc = parser.parse(config);
         root = doc.getDocumentElement();
         root.normalize();
      }
      catch (Exception e)
      {
         log.error(e);
         throw new ConfigurationException("Could not parse the config file");
      }
   }

   /**
    * Default schema validation error handler, that only logs validation exceptions.
    */
   private static class LoggingErrorHandler implements ErrorHandler
   {
      public void warning(SAXParseException exception) throws SAXException
      {
         log.warn("Configuration warning: " + exception.getMessage());
      }

      public void error(SAXParseException exception) throws SAXException
      {
         log.warn("Configuration error: " + exception.getMessage());
      }

      public void fatalError(SAXParseException exception) throws SAXException
      {
         log.warn("Configuration fatal error : " + exception.getMessage());
      }
   }
}
