package org.jboss.cache.config.parsing.element;

import org.jboss.cache.RegionManagerImpl;
import org.jboss.cache.config.*;
import org.jboss.cache.config.parsing.ParsedAttributes;
import org.jboss.cache.config.parsing.XmlConfigHelper;
import org.jboss.cache.config.parsing.XmlParserBase;
import org.jboss.cache.eviction.EvictionPolicy;
import org.jboss.cache.util.Util;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.util.ArrayList;
import java.util.List;

/**
 * Knows how to parse the <b>eviction</b> xml element.
 * <pre>
 * Note: class does not rely on element position in the configuration file.
 *       It does not rely on element's name either.
 * </pre>
 *
 * @author Mircea.Markus@jboss.com
 * @since 3.0
 */
public class EvictionElementParser extends XmlParserBase
{
   public EvictionConfig parseEvictionElement(Element evictionElement)
   {
      EvictionConfig evictionConfig = new EvictionConfig();
      String wakeUpInterval = getAttributeValue(evictionElement, "wakeUpInterval");
      if (existsAttribute(wakeUpInterval))
      {
         evictionConfig.setWakeupInterval(getInt(wakeUpInterval));
      }
      else
      {
       throw new ConfigurationException("Missing mandatory attribute wakeUpInterval"); 
      }

      String defaultPolicyClassStr = getAttributeValue(evictionElement, "defaultPolicyClass");
      if (existsAttribute(defaultPolicyClassStr)) evictionConfig.setDefaultEvictionPolicyClass(defaultPolicyClassStr);
      String defaultEventQueueSize = getAttributeValue(evictionElement, "defaultEventQueueSize");
      if (existsAttribute(defaultEventQueueSize)) evictionConfig.setDefaultEventQueueSize(getInt(defaultEventQueueSize));

      List<EvictionRegionConfig> evictionRegionConfigs = new ArrayList<EvictionRegionConfig>(3);
      Element rootRegion = getSingleElement("default", evictionElement);
      String defaultPolicyClass = evictionConfig.getDefaultEvictionPolicyClass();
      int defaultEvQueueSize = evictionConfig.getDefaultEventQueueSize();
      if (rootRegion != null)
      {
         EvictionRegionConfig erc = getEvictionRegionConfig(rootRegion, defaultPolicyClass, defaultEvQueueSize);
         erc.setRegionName(RegionManagerImpl.DEFAULT_REGION.toString());
         evictionRegionConfigs.add(erc);
      }

      NodeList regions = evictionElement.getElementsByTagName("region");
      for (int i = 0; i < regions.getLength(); i++)
      {
         Element regionConfig = (Element) regions.item(i);
         EvictionRegionConfig erc = getEvictionRegionConfig(regionConfig, defaultPolicyClass, defaultEvQueueSize);
         evictionRegionConfigs.add(erc);
      }
      evictionConfig.setEvictionRegionConfigs(evictionRegionConfigs);
      return evictionConfig;
   }

   private EvictionRegionConfig getEvictionRegionConfig(Element element, String defaultPolicyClass, int defaultEventQueueSize)
   {
      EvictionRegionConfig erc = new EvictionRegionConfig();
      erc.setRegionName(getAttributeValue(element, "name"));
      String queueSize = getAttributeValue(element, "eventQueueSize");
      if (existsAttribute(queueSize))
      {
         erc.setEventQueueSize(getInt(queueSize));
      }
      else
      {
         erc.setEventQueueSize(defaultEventQueueSize);
      }
      String policyClass = getAttributeValue(element, "policyClass");
      if (!existsAttribute(policyClass))
      {
         if (defaultPolicyClass == null)
         {
            throw new MissingPolicyException("There is no Eviction Policy Class specified on the region or for the entire cache!");
         }
         else
         {
            policyClass = defaultPolicyClass;
         }
      }
      EvictionPolicy policy;
      try
      {
         policy = (EvictionPolicy) Util.loadClass(policyClass).newInstance();
      }
      catch (Exception e)
      {
         throw new RuntimeException("Eviction class is not properly loaded in classloader", e);
      }

      EvictionPolicyConfig epc;
      try
      {
         epc = policy.getEvictionConfigurationClass().newInstance();
      }
      catch (Exception e)
      {
         throw new RuntimeException("Failed to instantiate eviction configuration of class " +
               policy.getEvictionConfigurationClass(), e);
      }
      parseEvictionPolicyConfig(element, epc);
      erc.setEvictionPolicyConfig(epc);
      return erc;
   }

   public static void parseEvictionPolicyConfig(Element element, EvictionPolicyConfig target)
   {
      target.reset();
      ParsedAttributes attributes = XmlConfigHelper.extractAttributes(element);
      XmlConfigHelper.setValues(target, attributes.stringAttribs, false, true);
      XmlConfigHelper.setValues(target, attributes.xmlAttribs, true, true);
      target.validate();
   }
}
