package org.jboss.cache.config.parsing.element;

import org.jboss.cache.config.CacheLoaderConfig;
import org.jboss.cache.config.ConfigurationException;
import org.jboss.cache.config.parsing.XmlConfigHelper;
import org.jboss.cache.config.parsing.XmlParserBase;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.util.Properties;

/**
 * Utility class for parsing the 'loaders' element in the .xml configuration file.
 * <pre>
 * Note: class does not rely on element position in the configuration file.
 *       It does not rely on element's name either.
 * </pre>
 *
 * @author Mircea.Markus@jboss.com
 * @since 3.0
 */
public class LoadersElementParser extends XmlParserBase
{
   public CacheLoaderConfig parseLoadersElement(Element element)
   {
      CacheLoaderConfig cacheLoaderConfig = new CacheLoaderConfig();
      String passivation = getAttributeValue(element, "passivation");
      if (existsAttribute(passivation)) cacheLoaderConfig.setPassivation(getBoolean(passivation));
      String shared = getAttributeValue(element, "shared");
      if (existsAttribute(shared)) cacheLoaderConfig.setShared(getBoolean(shared));
      String preload = getPreloadString(getSingleElement("preload", element));
      if (preload != null) cacheLoaderConfig.setPreload(preload);

      NodeList cacheLoaderNodes = element.getElementsByTagName("loader");
      for (int i = 0; i < cacheLoaderNodes.getLength(); i++)
      {
         Element indivElement = (Element) cacheLoaderNodes.item(i);
         CacheLoaderConfig.IndividualCacheLoaderConfig iclc = parseIndividualCacheLoaderConfig(indivElement);
         cacheLoaderConfig.addIndividualCacheLoaderConfig(iclc);
      }
      return cacheLoaderConfig;
   }

   private CacheLoaderConfig.IndividualCacheLoaderConfig parseIndividualCacheLoaderConfig(Element indivElement)
   {
      CacheLoaderConfig.IndividualCacheLoaderConfig iclc = new CacheLoaderConfig.IndividualCacheLoaderConfig();

      String async = getAttributeValue(indivElement, "async");
      if (existsAttribute(async)) iclc.setAsync(getBoolean(async));
      String fetchPersistentState = getAttributeValue(indivElement, "fetchPersistentState");
      if (existsAttribute(fetchPersistentState)) iclc.setFetchPersistentState(getBoolean(fetchPersistentState));
      String ignoreModifications = getAttributeValue(indivElement, "ignoreModifications");
      if (existsAttribute(ignoreModifications)) iclc.setIgnoreModifications(getBoolean(ignoreModifications));
      String purgeOnStartup = getAttributeValue(indivElement, "purgeOnStartup");
      if (existsAttribute(purgeOnStartup)) iclc.setPurgeOnStartup(getBoolean(purgeOnStartup));
      String clClass = getAttributeValue(indivElement, "class");
      if (!existsAttribute(clClass))
         throw new ConfigurationException("Missing 'class'  attribute for cache loader configuration");
      iclc.setClassName(clClass);
      iclc.setProperties(XmlConfigHelper.readPropertiesContents(indivElement, "properties"));
      CacheLoaderConfig.IndividualCacheLoaderConfig.SingletonStoreConfig ssc = parseSingletonStoreConfig(getSingleElement("singletonStore", indivElement));
      if (ssc != null)
      {
         iclc.setSingletonStoreConfig(ssc);
      }
      return iclc;
   }

   private String getPreloadString(Element preloadElement)
   {
      if (preloadElement == null) return null; //might be no preload
      NodeList nodesToPreload = preloadElement.getElementsByTagName("node");
      StringBuilder result = new StringBuilder();
      for (int i = 0; i < nodesToPreload.getLength(); i++)
      {
         Element node = (Element) nodesToPreload.item(i);
         String fqn2preload = getAttributeValue(node, "fqn");
         if (!existsAttribute(fqn2preload))
            throw new ConfigurationException("Missing 'fqn' attribute in 'preload\\norde' tag");
         if (i > 0) result.append(",");
         result.append(fqn2preload);
      }
      //no elements defined for preload so by default load the root
      if (nodesToPreload.getLength() == 0)
      {
         result.append("/");
      }
      return result.toString();
   }

   public CacheLoaderConfig.IndividualCacheLoaderConfig.SingletonStoreConfig parseSingletonStoreConfig(Element element)
   {
      if (element == null) return null; //might happen, this config option is not mandatory
      boolean singletonStoreEnabled = getBoolean(getAttributeValue(element, "enabled"));
      String singletonStoreClass = getAttributeValue(element, "class");
      CacheLoaderConfig.IndividualCacheLoaderConfig.SingletonStoreConfig ssc = new CacheLoaderConfig.IndividualCacheLoaderConfig.SingletonStoreConfig();
      if (existsAttribute(singletonStoreClass)) ssc.setSingletonStoreClass(singletonStoreClass);
      Properties singletonStoreproperties = XmlConfigHelper.readPropertiesContents(element, "properties");
      ssc.setSingletonStoreEnabled(singletonStoreEnabled);
      ssc.setSingletonStoreClass(singletonStoreClass);
      ssc.setSingletonStoreproperties(singletonStoreproperties);
      return ssc;
   }
}
