package org.jboss.cache.eviction;

import org.jboss.cache.config.ConfigurationComponent;
import org.jboss.cache.config.ConfigurationException;
import org.jboss.cache.config.Dynamic;
import org.jboss.cache.config.EvictionPolicyConfig;
import org.jboss.cache.util.Util;

/**
 * Base implementation of {@link EvictionPolicyConfig}. Adds properties
 * for the most commonly used config elements.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani</a>
 */
public abstract class EvictionPolicyConfigBase
      extends ConfigurationComponent
      implements EvictionPolicyConfig
{
   /**
    * The serialVersionUID
    */
   private static final long serialVersionUID = 4591691674370188932L;

   private String evictionPolicyClass;
   @Dynamic
   private int maxNodes = 0;

   @Deprecated
   @Dynamic
   private int minTimeToLiveSeconds = 0;


   /** value expressed in millis */
   @Dynamic
   private long minTimeToLive;

   /**
    * Can only be instantiated by a subclass.
    * <p/>
    * Calls {@link #setEvictionPolicyClassName()}.
    */
   protected EvictionPolicyConfigBase()
   {
      setEvictionPolicyClassName();
   }

   public String getEvictionPolicyClass()
   {
      return evictionPolicyClass;
   }

   public void setEvictionPolicyClass(String evictionPolicyClass)
   {
      testImmutability("evictionPolicyClass");
      this.evictionPolicyClass = evictionPolicyClass;
   }

   public int getMaxNodes()
   {
      return maxNodes;
   }

   public void setMaxNodes(int maxNodes)
   {
      testImmutability("maxNodes");
      this.maxNodes = maxNodes;
   }

   /** value expressed in millis */
   public long getMinTimeToLive()
   {
      return minTimeToLive;
   }

   /** value expressed in millis */
   public void setMinTimeToLive(long minTimeToLive)
   {
      this.minTimeToLive = minTimeToLive;
      this.minTimeToLiveSeconds = (int)(minTimeToLive / 1000);
   }

   /**
    * Use {@link #getMinTimeToLive()}
    */
   @Deprecated
   public int getMinTimeToLiveSeconds()
   {
      return this.minTimeToLiveSeconds;
   }

   /**
    * Use {@link #setMinTimeToLive(long)}
    */
   @Deprecated
   public void setMinTimeToLiveSeconds(int minTimeToLiveSeconds)
   {
      this.minTimeToLiveSeconds = minTimeToLiveSeconds;
      minTimeToLive = minTimeToLiveSeconds * 1000;
   }

   public void validate() throws ConfigurationException
   {
      // no-op
   }

   public boolean equals(Object o)
   {
      if (this == o) return true;
      if (!(o instanceof EvictionPolicyConfigBase)) return false;

      EvictionPolicyConfigBase that = (EvictionPolicyConfigBase) o;

      if (maxNodes != that.maxNodes) return false;
      if (minTimeToLive != that.minTimeToLive) return false;
      if (evictionPolicyClass != null ? !evictionPolicyClass.equals(that.evictionPolicyClass) : that.evictionPolicyClass != null)
         return false;

      return true;
   }

   public int hashCode()
   {
      int result;
      result = (evictionPolicyClass != null ? evictionPolicyClass.hashCode() : 0);
      result = 31 * result + maxNodes;
      result = 31 * result + minTimeToLiveSeconds;
      result = 31 * result + (int) (minTimeToLive ^ (minTimeToLive >>> 32));
      return result;
   }

   public void reset()
   {
      setEvictionPolicyClass(null);
      setMaxNodes(0);
      setMinTimeToLiveSeconds(0);
      setMinTimeToLive(0);
      setEvictionPolicyClassName();
   }

   /**
    * This method should be overridden by subclass implementers to set the default
    * {@link #setEvictionPolicyClass(String) policy class name} for the subclass.
    * This will be called when the implementation is constructed, but is also
    * called in {@link #reset()}.
    */
   abstract protected void setEvictionPolicyClassName();
}
