/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 * Created on March 25 2003
 */
package org.jboss.cache.eviction;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.Region;
import org.jboss.cache.util.concurrent.ConcurrentHashSet;

import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Timer threads to do periodic node clean up by running the eviction policy.
 *
 * @author Ben Wang 2-2004
 * @author Daniel Huang (dhuang@jboss.org)
 * @version $Revision: 6480 $
 */
public class EvictionTimerTask
{
   private Log log = LogFactory.getLog(EvictionTimerTask.class);

   private final Set<Region> processedRegions;
   private static AtomicInteger tcount = new AtomicInteger();
   private long wakeupInterval;
   ScheduledExecutorService scheduledExecutor;

   public EvictionTimerTask()
   {
      // synchronized set because we need to maintain thread safety
      // for dynamic configuration purposes.
      processedRegions = new ConcurrentHashSet<Region>();
   }

   public void init(long wakeupInterval)
   {
      if (log.isTraceEnabled())
         log.trace("Creating a new eviction listener with wakeupInterval millis set at " + wakeupInterval);
      this.wakeupInterval = wakeupInterval;
      start();
   }

   /**
    * Add a MarshRegion to process by the Eviction Thread.
    *
    * @param region MarshRegion to process.
    */
   public void addRegionToProcess(Region region)
   {
      processedRegions.add(region);
   }

   /**
    * Remove a MarshRegion to process from the Eviction thread.
    *
    * @param region
    */
   public void removeRegionToProcess(Region region)
   {
      processedRegions.remove(region);
   }

   public boolean isRegionRegisteredForProcessing(Region region)
   {
      return processedRegions.contains(region);
   }

   public Set<Region> getProcessedRegions()
   {
      return processedRegions;
   }

   public void stop()
   {
      if (log.isDebugEnabled()) log.debug("Stopping eviction timer");

      if (scheduledExecutor != null)
      {
         scheduledExecutor.shutdownNow();
      }
      scheduledExecutor = null;
   }

   private void start()
   {
      if (wakeupInterval < 1)
      {
         if (log.isInfoEnabled())
            log.info("Wakeup Interval set to " + wakeupInterval + ".  Not starting an eviction thread!");
         return;
      }
      scheduledExecutor = Executors.newSingleThreadScheduledExecutor(new ThreadFactory()
      {
         public Thread newThread(Runnable r)
         {
            return new Thread(r, "EvictionTimer-" + tcount.getAndIncrement());
         }
      });
      scheduledExecutor.scheduleWithFixedDelay(new Runnable()
      {
         public void run()
         {
            // Run the eviction thread.
            // This thread will synchronize the set of regions and iterate through every MarshRegion registered w/ the
            // Eviction thread. It also synchronizes on each individual region as it is being processed.
            processRegions();
         }
      }, wakeupInterval, wakeupInterval, TimeUnit.MILLISECONDS);
   }

   private void processRegions()
   {
      synchronized (processedRegions)
      {
         for (Region region : processedRegions)
         {
            handleRegion(region);
         }
      }
   }

   private void handleRegion(Region region)
   {
      synchronized (region)
      {
         final EvictionPolicy policy = region.getEvictionPolicy();
         final EvictionAlgorithm algo = policy.getEvictionAlgorithm();
         if (algo == null)
            throw new NullPointerException("algorithm null");
         try
         {
            algo.process(region);
         }
         catch (EvictionException e)
         {
            log.error("run(): error processing eviction with exception: " + e.toString()
                  + " will reset the eviction queue list.");
            region.resetEvictionQueues();
            log.debug("trace", e);
         }
      }
   }
}


