/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.eviction;

import org.jboss.cache.config.ConfigurationException;
import org.jboss.cache.config.Dynamic;

/**
 * Configuration implementation for {@link LRUPolicy}.
 * <p/>
 * If configured via XML, expects the following:
 * <p/>
 * <pre>
 * <region name="/maxAgeTest/">
 *    <attribute name="maxNodes">10000</attribute>
 *    <attribute name="timeToLiveSeconds">8</attribute>
 *    <attribute name="maxAgeSeconds">10</attribute>
 * </region>
 * </pre>
 *
 * @author Daniel Huang (dhuang@jboss.org)
 * @version $Revision: 6252 $
 */
public class LRUConfiguration extends EvictionPolicyConfigBase
{
   /**
    * The serialVersionUID
    */
   private static final long serialVersionUID = -3426716488271559729L;

   @Dynamic
   @Deprecated
   private int timeToLiveSeconds;
   @Dynamic
   @Deprecated
   private int maxAgeSeconds;


   /** value expressed in millis*/
   @Dynamic
   private long timeToLive;

   /** value expressed in millis*/
   @Dynamic
   private long maxAge;

   public LRUConfiguration()
   {
      super();
      // Force config of ttls
      setTimeToLive(-1);
   }

   public LRUConfiguration(long timeToLive, long maxAge)
   {
      super();
      setTimeToLive(timeToLive);
      setMaxAge(maxAge);
   }

   @Override
   protected void setEvictionPolicyClassName()
   {
      setEvictionPolicyClass(LRUPolicy.class.getName());
   }

   /** value expressed in millis*/
   public long getTimeToLive()
   {
      return timeToLive;
   }

   /** value expressed in millis*/
   public void setTimeToLive(long timeToLive)
   {
      testImmutability("timeToLive");
      this.timeToLive = timeToLive;
      this.timeToLiveSeconds = (int)(timeToLive/1000);
   }

   /** value expressed in millis*/
   public long getMaxAge()
   {
      return maxAge;
   }

   /** value expressed in millis*/
   public void setMaxAge(long maxAge)
   {
      testImmutability("maxAge");
      this.maxAge = maxAge;
      this.maxAgeSeconds = (int) (maxAge/1000);
   }

   /**
    * use {@link #getTimeToLive()}
    * @return
    */
   @Deprecated
   public int getTimeToLiveSeconds()
   {
      return timeToLiveSeconds;
   }

   /**
    * Use {@link #setTimeToLive(long)}
    */
   @Deprecated
   public void setTimeToLiveSeconds(int timeToLiveSeconds)
   {
      testImmutability("timeToLiveSeconds");
      this.timeToLiveSeconds = timeToLiveSeconds;
      timeToLive = timeToLiveSeconds * 1000;
   }

   /**
    * Use {@link #getMaxAge()}
    * @deprecated 
    */
   @Deprecated
   public int getMaxAgeSeconds()
   {
      return maxAgeSeconds;
   }

   /**
    * Use {@link #getMaxAge()}
    * @param maxAgeSeconds
    */
   @Deprecated
   public void setMaxAgeSeconds(int maxAgeSeconds)
   {
      testImmutability("maxAgeSeconds");
      this.maxAgeSeconds = maxAgeSeconds;
      this.maxAge = maxAgeSeconds * 1000;
   }

   /**
    * Requires a positive timeToLive value or ConfigurationException
    * is thrown.
    */
   @Override
   public void validate() throws ConfigurationException
   {
      if (timeToLive < 0)
      {
         throw new ConfigurationException("timeToLive must be " +
               "configured to a value greater than or equal to 0 for " + getEvictionPolicyClass());
      }
   }


   public String toString()
   {
      return "LRUConfiguration{" +
            "timeToLiveSeconds=" + timeToLiveSeconds +
            ", timeToLive=" + timeToLive +
            ", maxAge=" + maxAge +
            '}';
   }

   public boolean equals(Object o)
   {
      if (this == o) return true;
      if (!(o instanceof LRUConfiguration)) return false;
      if (!super.equals(o)) return false;

      LRUConfiguration that = (LRUConfiguration) o;

      if (maxAge != that.maxAge) return false;
      if (maxAgeSeconds != that.maxAgeSeconds) return false;
      if (timeToLive != that.timeToLive) return false;
      if (timeToLiveSeconds != that.timeToLiveSeconds) return false;

      return true;
   }

   public int hashCode()
   {
      int result = super.hashCode();
      result = 31 * result + timeToLiveSeconds;
      result = 31 * result + maxAgeSeconds;
      result = 31 * result + (int) (timeToLive ^ (timeToLive >>> 32));
      result = 31 * result + (int) (maxAge ^ (maxAge >>> 32));
      return result;
   }

   @Override
   public void reset()
   {
      super.reset();
      setTimeToLive(-1);
   }

   @Override
   public LRUConfiguration clone() throws CloneNotSupportedException
   {
      return (LRUConfiguration) super.clone();
   }

}
