/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.invocation;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.commands.VisitableCommand;
import org.jboss.cache.config.Option;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionContext;
import org.jboss.cache.transaction.TransactionTable;
import org.jboss.cache.util.ImmutableListCopy;

import javax.transaction.Transaction;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;

/**
 * This context holds information specific to a method invocation.  This is used for Optimistic and Pessimisic Node Locking Schemes.
 *
 * @author <a href="mailto:manik@jboss.org">Manik Surtani (manik@jboss.org)</a>
 */
@SuppressWarnings("deprecation")
public abstract class AbstractInvocationContext implements InvocationContext
{
   private static final Log log = LogFactory.getLog(AbstractInvocationContext.class);
   private static final boolean trace = log.isTraceEnabled();

   private Transaction transaction;
   private GlobalTransaction globalTransaction;
   protected TransactionContext transactionContext;
   private Option optionOverrides;
   // defaults to true.
   private boolean originLocal = true;
   private boolean localRollbackOnly;
   @Deprecated
   private MethodCall methodCall;
   @Deprecated
   private VisitableCommand command;

   /**
    * LinkedHashSet of locks acquired by the invocation. We use a LinkedHashSet because we need efficient Set semantics
    * but also need guaranteed ordering for use by lock release code (see JBCCACHE-874).
    * <p/>
    * This needs to be unchecked since we support both MVCC (Fqns held here) or legacy Opt/Pess locking (NodeLocks held here).
    * once we drop support for opt/pess locks we can genericise this to contain Fqns. - Manik Surtani, June 2008
    */
   protected LinkedHashSet invocationLocks;

   public void setLocalRollbackOnly(boolean localRollbackOnly)
   {
      this.localRollbackOnly = localRollbackOnly;
   }

   public Transaction getTransaction()
   {
      return transaction;
   }

   public void setTransaction(Transaction transaction)
   {
      this.transaction = transaction;
   }

   public TransactionContext getTransactionContext()
   {
      return transactionContext;
   }

   public void setTransactionContext(TransactionContext transactionContext)
   {
      this.transactionContext = transactionContext;
   }

   public GlobalTransaction getGlobalTransaction()
   {
      return globalTransaction;
   }

   public void setGlobalTransaction(GlobalTransaction globalTransaction)
   {
      this.globalTransaction = globalTransaction;
   }

   public Option getOptionOverrides()
   {
      if (optionOverrides == null)
      {
         optionOverrides = new Option();
      }
      return optionOverrides;
   }

   public boolean isOptionsUninitialised()
   {
      return optionOverrides == null;
   }

   public void setOptionOverrides(Option optionOverrides)
   {
      this.optionOverrides = optionOverrides;
   }

   public boolean isOriginLocal()
   {
      return originLocal;
   }

   @SuppressWarnings("unchecked")
   public List getLocks()
   {
      // first check transactional scope
      if (transactionContext != null) return transactionContext.getLocks();
      return invocationLocks == null || invocationLocks.isEmpty() ? Collections.emptyList() : new ImmutableListCopy(invocationLocks);
   }

   @SuppressWarnings("unchecked")
   public void addAllLocks(List locks)
   {
      // first check transactional scope
      if (transactionContext != null)
      {
         transactionContext.addAllLocks(locks);
      }
      else
      {
         // no need to worry about concurrency here - a context is only valid for a single thread.
         if (invocationLocks == null) invocationLocks = new LinkedHashSet(4);
         invocationLocks.addAll(locks);
      }
   }

   @SuppressWarnings("unchecked")
   public void addLock(Object lock)
   {
      // first check transactional scope
      if (transactionContext != null)
      {
         transactionContext.addLock(lock);
      }
      else
      {
         // no need to worry about concurrency here - a context is only valid for a single thread.
         if (invocationLocks == null) invocationLocks = new LinkedHashSet(4);
         invocationLocks.add(lock);
      }
   }

   @SuppressWarnings("unchecked")
   public void removeLock(Object lock)
   {
      // first check transactional scope
      if (transactionContext != null)
      {
         transactionContext.removeLock(lock);
      }
      else
      {
         // no need to worry about concurrency here - a context is only valid for a single thread.
         if (invocationLocks != null) invocationLocks.remove(lock);
      }
   }

   public void clearLocks()
   {
      // first check transactional scope
      if (transactionContext != null)
      {
         transactionContext.clearLocks();
      }
      else
      {
         // no need to worry about concurrency here - a context is only valid for a single thread.
         if (invocationLocks != null) invocationLocks.clear();
      }
   }

   public boolean hasLock(Object lock)
   {
      // first check transactional scope
      if (transactionContext != null)
      {
         return transactionContext.hasLock(lock);
      }
      else
      {
         return invocationLocks != null && invocationLocks.contains(lock);
      }
   }

   public boolean isLockingSuppressed()
   {
      return getOptionOverrides() != null && getOptionOverrides().isSuppressLocking();
   }

   public void setOriginLocal(boolean originLocal)
   {
      this.originLocal = originLocal;
   }

   @Override
   public String toString()
   {
      return "InvocationContext{" +
            "transaction=" + transaction +
            ", globalTransaction=" + globalTransaction +
            ", transactionContext=" + transactionContext +
            ", optionOverrides=" + optionOverrides +
            ", originLocal=" + originLocal +
            '}';
   }

   public boolean isLocalRollbackOnly()
   {
      return localRollbackOnly;
   }

   public void reset()
   {
      transaction = null;
      globalTransaction = null;
      optionOverrides = null;
      originLocal = true;
      invocationLocks = null;
      methodCall = null;
      command = null;
   }

   @SuppressWarnings("unchecked")
   protected void doCopy(AbstractInvocationContext copy)
   {
      copy.command = command;
      copy.globalTransaction = globalTransaction;
      copy.invocationLocks = invocationLocks == null ? null : new LinkedHashSet(invocationLocks);
      copy.localRollbackOnly = localRollbackOnly;
      copy.optionOverrides = optionOverrides == null ? null : optionOverrides.copy();
      copy.originLocal = originLocal;
      copy.transaction = transaction;
      copy.transactionContext = transactionContext;
   }

   public void setState(InvocationContext template)
   {
      if (template == null)
      {
         throw new NullPointerException("Template InvocationContext passed in to InvocationContext.setState() passed in is null");
      }

      this.setGlobalTransaction(template.getGlobalTransaction());
      this.setLocalRollbackOnly(template.isLocalRollbackOnly());
      this.setOptionOverrides(template.getOptionOverrides());
      this.setOriginLocal(template.isOriginLocal());
      this.setTransaction(template.getTransaction());
   }

   @Override
   public boolean equals(Object o)
   {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;

      final AbstractInvocationContext that = (AbstractInvocationContext) o;

      if (localRollbackOnly != that.localRollbackOnly) return false;
      if (originLocal != that.originLocal) return false;
      if (globalTransaction != null ? !globalTransaction.equals(that.globalTransaction) : that.globalTransaction != null)
      {
         return false;
      }
      if (optionOverrides != null ? !optionOverrides.equals(that.optionOverrides) : that.optionOverrides != null)
      {
         return false;
      }
      if (transaction != null ? !transaction.equals(that.transaction) : that.transaction != null) return false;

      return true;
   }

   @Override
   public int hashCode()
   {
      int result;
      result = (transaction != null ? transaction.hashCode() : 0);
      result = 29 * result + (globalTransaction != null ? globalTransaction.hashCode() : 0);
      result = 29 * result + (optionOverrides != null ? optionOverrides.hashCode() : 0);
      result = 29 * result + (originLocal ? 1 : 0);
      result = 29 * result + (localRollbackOnly ? 1 : 0);
      return result;
   }

   @Deprecated
   @SuppressWarnings("deprecation")
   public MethodCall getMethodCall()
   {
      if (methodCall == null)
      {
         methodCall = createMethodCall();
      }
      return methodCall;
   }

   @SuppressWarnings("deprecation")
   private MethodCall createMethodCall()
   {
      if (command == null) return null;
      MethodCall call = new MethodCall();
      call.setMethodId(command.getCommandId());
      call.setArgs(command.getParameters());
      return call;
   }

   @Deprecated
   public void setMethodCall(MethodCall methodCall)
   {
      this.methodCall = methodCall;
   }

   public long getLockAcquisitionTimeout(long timeout)
   {
      // TODO: this stuff really doesn't belong here.  Put it somewhere else.
      if (getOptionOverrides() != null
            && getOptionOverrides().getLockAcquisitionTimeout() >= 0)
      {
         timeout = getOptionOverrides().getLockAcquisitionTimeout();
      }
      return timeout;
   }

   @Deprecated
   @SuppressWarnings("deprecation")
   public void setCommand(VisitableCommand cacheCommand)
   {
      this.command = cacheCommand;
   }

   @Deprecated
   @SuppressWarnings("deprecation")
   public VisitableCommand getCommand()
   {
      return command;
   }

   public boolean isValidTransaction()
   {
      // ought to move to the transaction context
      return transaction != null && TransactionTable.isValid(transaction);
   }

   public void throwIfNeeded(Throwable e) throws Throwable
   {
      // TODO: this stuff really doesn't belong here.  Put it somewhere else.
      Option optionOverride = getOptionOverrides();
      boolean shouldRethtrow = optionOverride == null || !optionOverride.isFailSilently();
      if (!shouldRethtrow)
      {
         if (trace)
            log.trace("There was a problem handling this request, but failSilently was set, so suppressing exception", e);
         return;
      }
      throw e;
   }
}
