package org.jboss.cache.invocation;

import org.jboss.cache.Fqn;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.commands.VisitableCommand;
import org.jboss.cache.config.Option;
import org.jboss.cache.marshall.MethodCall;
import org.jboss.cache.transaction.GlobalTransaction;
import org.jboss.cache.transaction.TransactionContext;

import javax.transaction.Transaction;
import java.util.List;
import java.util.Map;

/**
 * A context that holds information regarding the scope of a single invocation.  May delegate some calls to a {@link org.jboss.cache.transaction.TransactionContext}
 * if one is in scope.
 * <p/>
 * Note that prior to 3.0.0, InvocationContext was a concrete class and not an interface.
 * <p/>
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @see org.jboss.cache.transaction.TransactionContext
 */
@SuppressWarnings("deprecation")
public interface InvocationContext
{
   /**
    * Retrieves a node from the registry of looked up nodes in the current scope.
    *
    * @param fqn fqn to look up
    * @return a node, or null if it cannot be found.
    * @since 3.0.
    */
   NodeSPI lookUpNode(Fqn fqn);

   /**
    * Puts an entry in the registry of looked up nodes in the current scope.
    *
    * @param f fqn to add
    * @param n node to add
    * @since 3.0.
    */
   void putLookedUpNode(Fqn f, NodeSPI n);

   /**
    * Clears the registry of looked up nodes.
    *
    * @since 3.0.
    */
   void clearLookedUpNodes();

   /**
    * Retrieves a map of nodes looked up within the current invocation's scope.
    *
    * @return a map of looked up nodes.
    * @since 3.0
    */
   Map<Fqn, NodeSPI> getLookedUpNodes();

   /**
    * Marks teh context as only rolling back.
    *
    * @param localRollbackOnly if true, the context is only rolling back.
    */
   void setLocalRollbackOnly(boolean localRollbackOnly);

   /**
    * Retrieves the transaction associated with this invocation
    *
    * @return The transaction associated with this invocation
    */
   Transaction getTransaction();

   /**
    * Sets a transaction object on the invocation context.
    *
    * @param transaction transaction to set
    */
   void setTransaction(Transaction transaction);

   /**
    * @return the transaction entry associated with the current transaction, or null if the current thread is not associated with a transaction.
    * @since 2.2.0
    */
   TransactionContext getTransactionContext();

   /**
    * Sets the transaction context to be associated with the current thread.
    *
    * @param transactionContext transaction context to set
    * @since 2.2.0
    */
   void setTransactionContext(TransactionContext transactionContext);

   /**
    * Retrieves the global transaction associated with this invocation
    *
    * @return the global transaction associated with this invocation
    */
   GlobalTransaction getGlobalTransaction();

   /**
    * Sets the global transaction associated with this invocation
    *
    * @param globalTransaction global transaction to set
    */
   void setGlobalTransaction(GlobalTransaction globalTransaction);

   /**
    * Retrieves the option overrides associated with this invocation
    *
    * @return the option overrides associated with this invocation
    */
   Option getOptionOverrides();

   /**
    * @return true of no options have been set on this context, false otherwise.
    */
   boolean isOptionsUninitialised();

   /**
    * Sets the option overrides to be associated with this invocation
    *
    * @param optionOverrides options to set
    */
   void setOptionOverrides(Option optionOverrides);

   /**
    * Tests if this invocation originated locally or from a remote cache.
    *
    * @return true if the invocation originated locally.
    */
   boolean isOriginLocal();

   /**
    * Returns an immutable,  defensive copy of the List of locks currently maintained for the current scope.
    * <p/>
    * Note that if a transaction is in scope, implementations should retrieve these locks from the {@link org.jboss.cache.transaction.TransactionContext}.
    * Retrieving locks from this method should always ensure they are retrieved from  the appropriate scope.
    * <p/>
    * Note that currently (as of 3.0.0) this list is unchecked.  This is to allow support for both MVCC (which uses Fqns as locks)
    * as well as legacy Optimistic and Pessimistic Locking schemes (which use {@link org.jboss.cache.lock.NodeLock} as locks).  Once support for
    * legacy node locking schemes are dropped, this method will be more strongly typed to return <tt>List<Fqn></tt>.
    *
    * @return locks held in current scope.
    */
   @SuppressWarnings("unchecked")
   List getLocks();

   /**
    * Adds a List of locks to the currently maintained collection of locks acquired.
    * <p/>
    * Note that if a transaction is in scope, implementations should record locks on the {@link org.jboss.cache.transaction.TransactionContext}.
    * Adding locks using this method should always ensure they are applied to the appropriate scope.
    * <p/>
    * Note that currently (as of 3.0.0) this list is unchecked.  This is to allow support for both MVCC (which uses Fqns as locks)
    * as well as legacy Optimistic and Pessimistic Locking schemes (which use {@link org.jboss.cache.lock.NodeLock} as locks).  Once support for
    * legacy node locking schemes are dropped, this method will be more strongly typed to accept <tt>List<Fqn></tt>.
    *
    * @param locks locks to add
    */
   @SuppressWarnings("unchecked")
   void addAllLocks(List locks);

   /**
    * Adds a lock to the currently maintained collection of locks acquired.
    * <p/>
    * Note that if a transaction is in scope, implementations should record this lock on the {@link org.jboss.cache.transaction.TransactionContext}.
    * Using this method should always ensure that the appropriate scope is used.
    * <p/>
    * Note that currently (as of 3.0.0) this lock is weakly typed.  This is to allow support for both MVCC (which uses {@link Fqn}s as locks)
    * as well as legacy Optimistic and Pessimistic Locking schemes (which use {@link org.jboss.cache.lock.NodeLock} as locks).  Once support for
    * legacy node locking schemes are dropped, this method will be more strongly typed to accept {@link Fqn}.
    *
    * @param lock lock to add
    */
   @SuppressWarnings("unchecked")
   void addLock(Object lock);

   /**
    * Removes a lock from the currently maintained collection of locks acquired.
    * <p/>
    * Note that if a transaction is in scope, implementations should remove this lock from the {@link org.jboss.cache.transaction.TransactionContext}.
    * Using this method should always ensure that the lock is removed from the appropriate scope.
    * <p/>
    * Note that currently (as of 3.0.0) this lock is weakly typed.  This is to allow support for both MVCC (which uses {@link Fqn}s as locks)
    * as well as legacy Optimistic and Pessimistic Locking schemes (which use {@link org.jboss.cache.lock.NodeLock} as locks).  Once support for
    * legacy node locking schemes are dropped, this method will be more strongly typed to accept {@link Fqn}.
    *
    * @param lock lock to remove
    */
   @SuppressWarnings("unchecked")
   void removeLock(Object lock);

   /**
    * Clears all locks from the currently maintained collection of locks acquired.
    * <p/>
    * Note that if a transaction is in scope, implementations should clear locks from the {@link org.jboss.cache.transaction.TransactionContext}.
    * Using this method should always ensure locks are cleared in the appropriate scope.
    * <p/>
    * Note that currently (as of 3.0.0) this lock is weakly typed.  This is to allow support for both MVCC (which uses {@link Fqn}s as locks)
    * as well as legacy Optimistic and Pessimistic Locking schemes (which use {@link org.jboss.cache.lock.NodeLock} as locks).  Once support for
    * legacy node locking schemes are dropped, this method will be more strongly typed to accept {@link Fqn}.
    */
   void clearLocks();


   /**
    * Note that if a transaction is in scope, implementations should test this lock from on {@link org.jboss.cache.transaction.TransactionContext}.
    * Using this method should always ensure locks checked in the appropriate scope.
    *
    * @param lock lock to test
    * @return true if the lock being tested is already held in the current scope, false otherwise.
    */
   boolean hasLock(Object lock);

   /**
    * @return true if options exist to suppress locking - false otherwise.  Note that this is only used by the {@link org.jboss.cache.interceptors.PessimisticLockInterceptor}.
    */
   boolean isLockingSuppressed();

   /**
    * If set to true, the invocation is assumed to have originated locally.  If set to false,
    * assumed to have originated from a remote cache.
    *
    * @param originLocal flag to set
    */
   void setOriginLocal(boolean originLocal);

   /**
    * @return true if the current transaction is set to rollback only.
    */
   boolean isLocalRollbackOnly();

   /**
    * Resets the context, freeing up any references.
    */
   void reset();

   /**
    * This is a "copy-factory-method" that should be used whenever a clone of this class is needed.  The resulting instance
    * is equal() to, but not ==, to the InvocationContext invoked on.  Note that this is a shallow copy with the exception
    * of the Option object, which is deep, as well as any collections held on the context such as locks.  Note that the reference
    * to a {@link org.jboss.cache.transaction.TransactionContext}, if any, is maintained.
    *
    * @return a new InvocationContext
    */
   @SuppressWarnings("unchecked")
   InvocationContext copy();

   /**
    * Sets the state of the InvocationContext based on the template context passed in
    *
    * @param template template to copy from
    */
   void setState(InvocationContext template);

   /**
    * @return the method call associated with this invocation
    */
   @Deprecated
   @SuppressWarnings("deprecation")
   MethodCall getMethodCall();


   /**
    * Sets the method call associated with this invocation.
    *
    * @param methodCall methodcall to set
    * @deprecated not used anymore.  Interceptors will get a {@link org.jboss.cache.commands.ReplicableCommand} instance passed in along with an InvocationContext.
    */
   @Deprecated
   void setMethodCall(MethodCall methodCall);

   /**
    * If the lock acquisition timeout is overridden for current call using an option, then return that one.
    * If not overridden, return default value.
    *
    * @param defaultTimeout timeout to fall back to
    * @return timeout to use
    */
   long getLockAcquisitionTimeout(long defaultTimeout);

   /**
    * This is only used for backward compatibility with old interceptors implementation and should <b>NOT</b> be
    * use by any new custom interceptors. The commands is now passed in as the second param in each implementing
    * handlers (handler = method in ChainedInterceptor class)
    *
    * @param cacheCommand command to set
    */
   @Deprecated
   @SuppressWarnings("deprecation")
   void setCommand(VisitableCommand cacheCommand);

   /**
    * @return command that is in scope
    * @see #setCommand(org.jboss.cache.commands.VisitableCommand)
    */
   @Deprecated
   @SuppressWarnings("deprecation")
   VisitableCommand getCommand();

   /**
    * @return true if there is current transaction associated with the invocation, and this transaction is in a valid state.
    */
   boolean isValidTransaction();

   /**
    * Throws the given throwable provided no options suppress or prevent this from happening.
    *
    * @param throwable throwable to throw
    * @throws Throwable if allowed to throw one.
    */
   void throwIfNeeded(Throwable throwable) throws Throwable;
}
