package org.jboss.cache.mvcc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.AbstractNodeFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.InternalNode;
import org.jboss.cache.NodeSPI;
import org.jboss.cache.UnversionedNode;
import org.jboss.cache.factories.annotations.Start;
import org.jboss.cache.invocation.InvocationContext;
import org.jboss.cache.lock.IsolationLevel;

import java.util.Map;

/**
 * Creates nodes specific to MVCC logic.
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 3.0
 */
public class MVCCNodeFactory<K, V> extends AbstractNodeFactory<K, V>
{
   private boolean useRepeatableRead;
   private static final NullMarkerNode NULL_MARKER = new NullMarkerNode(null);
   private static final Log log = LogFactory.getLog(MVCCNodeFactory.class);
   private static final boolean trace = log.isTraceEnabled();
   private boolean lockChildForInsertRemove;

   /**
    * Initialises the node factory with the configuration from the cache.
    */
   @Start
   public void init()
   {
      useRepeatableRead = configuration.getIsolationLevel() == IsolationLevel.REPEATABLE_READ;
      lockChildForInsertRemove = configuration.isLockParentForChildInsertRemove();
   }

   /**
    * Creates an MVCC wrapped node - either a {@link org.jboss.cache.mvcc.ReadCommittedNode} or it's subclass, a
    * {@link org.jboss.cache.mvcc.RepeatableReadNode} based on cache configuration.  If a null is passed in as the InternalNode,
    * this method will return a special {@link org.jboss.cache.mvcc.NullMarkerNode} instance if using repeatable read,
    * or a null if read committed.
    *
    * @param node internal node to wrap.
    * @return a ReadCommittedNode
    */
   @Override
   @SuppressWarnings("unchecked")
   public ReadCommittedNode createWrappedNode(InternalNode<K, V> node)
   {
      if (node == null) return useRepeatableRead ? NULL_MARKER : null;
      ReadCommittedNode rcn = useRepeatableRead ? new RepeatableReadNode(node) : new ReadCommittedNode(node);
      rcn.initialize(configuration, invocationContextContainer, componentRegistry, interceptorChain);
      rcn.injectDependencies(cache);
      return rcn;
   }

   @Override
   public NodeSPI<K, V> createNode(Fqn fqn, NodeSPI<K, V> parent, Map<K, V> data)
   {
      throw new UnsupportedOperationException();
   }

   @Override
   public NodeSPI<K, V> createNode(Fqn fqn, NodeSPI<K, V> parent)
   {
      throw new UnsupportedOperationException();
   }

   @Override
   @SuppressWarnings("unchecked")
   public NodeSPI<K, V> createRootNode()
   {
      return createWrappedNode(createInternalNode(Fqn.ROOT));
   }

   @Override
   public InternalNode<K, V> createInternalNode(Fqn fqn)
   {
      UnversionedNode<K, V> un = new UnversionedNode<K, V>(fqn, cache, lockChildForInsertRemove);
      return useRepeatableRead ? un : new NodeReference<K, V>(un);
   }

   @Override
   public NodeSPI<K, V> createNode(Object childName, NodeSPI<K, V> parent, Map<K, V> data)
   {
      return createNode(Fqn.fromRelativeElements(parent.getFqn(), childName), parent, data);
   }

   @Override
   public NodeSPI<K, V> createNode(Object childName, NodeSPI<K, V> parent)
   {
      return createNode(Fqn.fromRelativeElements(parent.getFqn(), childName), parent);
   }

   @Override
   @SuppressWarnings("unchecked")
   public InternalNode<K, V> createChildNode(Fqn fqn, InternalNode<K, V> parent, InvocationContext ctx, boolean attachToParent)
   {
      InternalNode<K, V> child;
      if (fqn == null) throw new IllegalArgumentException("null child fqn");

      child = dataContainer.peekInternalNode(fqn, false);

      if (child == null)
      {
         cache.getNotifier().notifyNodeCreated(fqn, true, ctx);
         InternalNode<K, V> newChild = createInternalNode(fqn);
         if (attachToParent) parent.addChild(newChild);
         // addChild actually succeeded!
         child = newChild;

         if (trace) log.trace("Created new child with fqn [" + fqn + "]");

         // notify if we actually created a new child
         cache.getNotifier().notifyNodeCreated(fqn, false, ctx);
      }
      return child;
   }
}
