package org.jboss.cache.mvcc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.DataContainer;
import org.jboss.cache.Fqn;
import org.jboss.cache.InternalNode;
import org.jboss.cache.invocation.InvocationContext;
import static org.jboss.cache.mvcc.ReadCommittedNode.Flags.CHANGED;
import static org.jboss.cache.mvcc.ReadCommittedNode.Flags.DELETED;
import org.jboss.cache.optimistic.DataVersioningException;

/**
 * A node delegate that encapsulates repeatable read semantics when writes are initiated, committed or rolled back.
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 3.0
 */
public class RepeatableReadNode extends ReadCommittedNode
{
   private static final Log log = LogFactory.getLog(RepeatableReadNode.class);

   public RepeatableReadNode(InternalNode node)
   {
      super(node);
   }

   @Override
   public void markForUpdate(DataContainer container, boolean writeSkewCheck)
   {
      if (isFlagSet(CHANGED)) return; // already copied

      Fqn fqn = getFqn();

      // mark node as changed.
      setFlag(CHANGED);

      if (writeSkewCheck)
      {
         // check for write skew.
         InternalNode underlyingNode = container.peekInternalNode(fqn, true);

         if (underlyingNode != null && underlyingNode != node)
         {
            String errormsg = new StringBuilder().append("Detected write skew on Fqn [").append(fqn).append("].  Another process has changed the node since we last read it!").toString();
            if (log.isWarnEnabled()) log.warn(errormsg + ".  Unable to copy node for update.");
            throw new DataVersioningException(errormsg);
         }
      }

      // make a backup copy
      backup = node;
      node = backup.copy();
   }

   @Override
   @SuppressWarnings("unchecked")
   protected void updateNode(InvocationContext ctx, DataContainer dataContainer)
   {
      Fqn fqn = getFqn();
      if (fqn.isRoot())
      {
         dataContainer.setRoot(node);
      }
      else if (!isFlagSet(DELETED))
      {
         InternalNode parent = lookupParent(fqn, ctx, dataContainer);
         parent.addChild(node);
      }
   }
}
