/*
 * JBoss, the OpenSource J2EE webOS
 * 
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.statetransfer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheException;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.InternalNode;
import org.jboss.cache.NodeFactory;
import org.jboss.cache.Region;
import org.jboss.cache.buddyreplication.BuddyManager;
import org.jboss.cache.eviction.EvictedEventNode;
import org.jboss.cache.eviction.NodeEventType;
import org.jboss.cache.factories.annotations.Inject;
import org.jboss.cache.factories.annotations.Start;
import org.jboss.cache.invocation.InvocationContext;
import org.jboss.cache.loader.CacheLoader;
import org.jboss.cache.loader.CacheLoaderManager;
import org.jboss.cache.marshall.NodeData;
import org.jboss.cache.marshall.NodeDataExceptionMarker;
import org.jboss.cache.marshall.NodeDataMarker;
import org.jboss.cache.notifications.event.NodeModifiedEvent;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class DefaultStateTransferIntegrator implements StateTransferIntegrator
{

   private static final Log log = LogFactory.getLog(DefaultStateTransferIntegrator.class);
   private static final boolean trace = log.isTraceEnabled();

   private CacheSPI cache;

   private NodeFactory factory;

   private Set<Fqn> internalFqns;


   @Inject
   public void inject(CacheSPI<?, ?> cache, NodeFactory nodefactory)
   {
      this.cache = cache;
      this.factory = nodefactory;
   }

   @Start(priority = 14)
   public void start()
   {
      this.internalFqns = cache.getInternalFqns();
   }

   public void integrateState(ObjectInputStream ois, Object target, Fqn targetRoot, boolean integratePersistentState) throws Exception
   {
      // pop version from the stream first!
      short version = (Short) cache.getMarshaller().objectFromObjectStream(ois);
      log.info("Using version " + version);
      integrateTransientState(ois, (InternalNode) target);
      if (trace) log.trace("Reading marker for nonexistent associated state");
      cache.getMarshaller().objectFromObjectStream(ois);
      if (integratePersistentState)
      {
         integratePersistentState(ois, targetRoot);
      }
   }

   protected void integrateTransientState(ObjectInputStream in, InternalNode target) throws Exception
   {
      boolean transientSet = false;
      try
      {
         if (trace)
         {
            log.trace("integrating transient state for " + target);
         }

         integrateTransientState(target, in);

         transientSet = true;

         if (trace)
         {
            log.trace("transient state successfully integrated");
         }

         notifyAllNodesCreated(cache.getInvocationContext(), target);
      }
      catch (Exception e)
      {
         throw new CacheException(e);
      }
      finally
      {
         if (!transientSet)
         {
            target.clear();
            target.removeChildren();
         }
      }
   }

   protected void integratePersistentState(ObjectInputStream in, Fqn targetFqn) throws Exception
   {
      CacheLoaderManager loaderManager = cache.getCacheLoaderManager();
      CacheLoader loader = loaderManager == null ? null : loaderManager.getCacheLoader();
      if (loader == null)
      {
         if (trace)
         {
            log.trace("cache loader is null, will not attempt to integrate persistent state");
         }
      }
      else
      {

         if (trace)
         {
            log.trace("integrating persistent state using " + loader.getClass().getName());
         }

         boolean persistentSet = false;
         try
         {
            if (targetFqn.isRoot())
            {
               loader.storeEntireState(in);
            }
            else
            {
               loader.storeState(targetFqn, in);
            }
            persistentSet = true;
         }
         catch (ClassCastException cce)
         {
            log.error("Failed integrating persistent state. One of cacheloaders is not"
                  + " adhering to state stream format. See JBCACHE-738.");
            throw cce;
         }
         finally
         {
            if (!persistentSet)
            {
               log.warn("persistent state integration failed, removing all nodes from loader");
               loader.remove(targetFqn);
            }
            else
            {
               if (trace)
               {
                  log.trace("persistent state integrated successfully");
               }
            }
         }
      }
   }

   /**
    * Generates NodeAdded notifications for all nodes of the tree. This is
    * called whenever the tree is initially retrieved (state transfer)
    */
   private void notifyAllNodesCreated(InvocationContext ctx, InternalNode curr)
   {
      if (curr == null) return;
      ctx.setOriginLocal(false);
      cache.getNotifier().notifyNodeCreated(curr.getFqn(), true, ctx);
      cache.getNotifier().notifyNodeCreated(curr.getFqn(), false, ctx);
      // AND notify that they have been modified!!
      if (!curr.getKeys().isEmpty())
      {
         cache.getNotifier().notifyNodeModified(curr.getFqn(), true, NodeModifiedEvent.ModificationType.PUT_MAP, Collections.emptyMap(), ctx);
         cache.getNotifier().notifyNodeModified(curr.getFqn(), false, NodeModifiedEvent.ModificationType.PUT_MAP, curr.getData(), ctx);
      }
      ctx.setOriginLocal(true);

      Set<InternalNode> children = curr.getChildren();
      for (InternalNode n : children) notifyAllNodesCreated(ctx, n);
   }

   private void integrateTransientState(InternalNode target, ObjectInputStream in) throws Exception
   {
      target.removeChildren();

      List<NodeData> list = readNodesAsList(in);
      if (list != null)
      {
         // if the list was null we read an EOF marker!!  So don't bother popping it off the stack later.
         Iterator<NodeData> nodeDataIterator = list.iterator();

         // Read the first NodeData and integrate into our target
         if (nodeDataIterator.hasNext())
         {
            NodeData nd = nodeDataIterator.next();

            //are there any transient nodes at all?
            if (nd != null && !nd.isMarker())
            {
               Map attributes = nd.getAttributes();

               target.setInternalState(attributes);

               // Check whether this is an integration into the buddy backup
               // subtree
               Fqn tferFqn = nd.getFqn();
               Fqn tgtFqn = target.getFqn();
               boolean move = tgtFqn.isChildOrEquals(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN)
                     && !tferFqn.isChildOrEquals(tgtFqn);
               // If it is an integration, calculate how many levels of offset
               int offset = move ? tgtFqn.size() - tferFqn.size() : 0;

               integrateStateTransferChildren(target, offset, nodeDataIterator);

               integrateRetainedNodes(target);
            }
         }

         // read marker off stack
//         cache.getMarshaller().objectFromObjectStream(in);
      }
   }

   @SuppressWarnings("unchecked")
   private List<NodeData> readNodesAsList(ObjectInputStream in) throws Exception
   {
      Object obj = cache.getMarshaller().objectFromObjectStream(in);
      if (obj instanceof NodeDataMarker) return null;

      return (List<NodeData>) obj;
   }

   private NodeData integrateStateTransferChildren(InternalNode parent, int offset, Iterator<NodeData> nodeDataIterator)
         throws IOException, ClassNotFoundException
   {
      int parent_level = parent.getFqn().size();
      int target_level = parent_level + 1;
      Fqn fqn;
      int size;
      NodeData nd = nodeDataIterator.hasNext() ? nodeDataIterator.next() : null;
      while (nd != null && !nd.isMarker())
      {
         fqn = nd.getFqn();
         // If we need to integrate into the buddy backup subtree,
         // change the Fqn to fit under it
         if (offset > 0)
         {
            fqn = Fqn.fromRelativeFqn(parent.getFqn().getAncestor(offset), fqn);
         }
         size = fqn.size();
         if (size <= parent_level)
         {
            return nd;
         }
         else if (size > target_level)
         {
            throw new IllegalStateException("NodeData " + fqn + " is not a direct child of " + parent.getFqn());
         }

         Map attrs = nd.getAttributes();

         InternalNode internalTarget = factory.createInternalNode(fqn);
         internalTarget.setInternalState(attrs);
         parent.addChild(internalTarget);

         // JBCACHE-913
         Region region = cache.getRegion(fqn, false);
         if (region != null && region.getEvictionPolicy() != null)
         {
            region.putNodeEvent(new EvictedEventNode(fqn, NodeEventType.ADD_NODE_EVENT,
                  attrs == null ? 0 : attrs.size()));
         }

         // Recursively call, which will walk down the tree
         // and return the next NodeData that's a child of our parent
         nd = integrateStateTransferChildren(internalTarget, offset, nodeDataIterator);
      }
      if (nd != null && nd.isExceptionMarker())
      {
         NodeDataExceptionMarker ndem = (NodeDataExceptionMarker) nd;
         throw new CacheException("State provider node " + ndem.getCacheNodeIdentity()
               + " threw exception during loadState", ndem.getCause());
      }
      return null;
   }

   private Set<InternalNode> retainInternalNodes(InternalNode target)
   {
      Set<InternalNode> result = new HashSet<InternalNode>();
      Fqn targetFqn = target.getFqn();
      for (Fqn internalFqn : internalFqns)
      {
         if (internalFqn.isChildOf(targetFqn))
         {
            InternalNode internalNode = getInternalNode(target, internalFqn);
            if (internalNode != null)
            {
               result.add(internalNode);
            }
         }
      }

      return result;
   }

   private InternalNode getInternalNode(InternalNode parent, Fqn internalFqn)
   {
      Object name = internalFqn.get(parent.getFqn().size());
      InternalNode result = parent.getChild(name);
      if (result != null)
      {
         if (internalFqn.size() < result.getFqn().size())
         {
            // need to recursively walk down the tree
            result = getInternalNode(result, internalFqn);
         }
      }
      return result;
   }

   private void integrateRetainedNodes(InternalNode target)
   {
      Set<InternalNode> retainedNodes = retainInternalNodes(target);
      Fqn rootFqn = target.getFqn();
      for (InternalNode retained : retainedNodes)
      {
         if (retained.getFqn().isChildOf(rootFqn))
         {
            integrateRetainedNode(target, retained);
         }
      }
   }

   private void integrateRetainedNode(InternalNode ancestor, InternalNode descendant)
   {
      Fqn descFqn = descendant.getFqn();
      Fqn ancFqn = ancestor.getFqn();
      Object name = descFqn.get(ancFqn.size());
      InternalNode child = ancestor.getChild(name);
      if (ancFqn.size() == descFqn.size() + 1)
      {
         if (child == null)
         {
            ancestor.addChild(name, descendant);
         }
         else
         {
            log.warn("Received unexpected internal node " + descFqn + " in transferred state");
         }
      }
      else
      {
         if (child == null)
         {
            // Missing level -- have to create empty node
            // This shouldn't really happen -- internal fqns should
            // be immediately under the root
//            child = factory.createInternalNode(name, ancestor);
//            ancestor.addChild(name, child);

            // since this shouldn't happen, deal with it. - Manik - Jul08
            throw new NullPointerException("Child is null");
         }

         // Keep walking down the tree
         integrateRetainedNode(child, descendant);
      }
   }
}
