package org.jboss.cache.util;

import net.jcip.annotations.Immutable;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.AbstractMap;
import java.util.Map;
import java.util.Set;

/**
 * Along the lines of ImmutableListCopy and ImmutableSetCopy.  This lightweight wrapper primarily meant for iteration.
 * performance of get() and containsKey() are not as efficient as that of a {@link java.util.HashMap}, but iteration is
 * fast since the entry table is a fixed size immutable array.
 * <p/>
 * Typically used in place of the common idiom:
 * <code>
 * return Collections.unmodifiableMap(new HashMap( myInternalMap ));
 * </code>
 *
 * @author Manik Surtani (<a href="mailto:manik@jboss.org">manik@jboss.org</a>)
 * @since 3.0
 */
@Immutable
public class ImmutableMapCopy<K, V> extends AbstractMap<K, V> implements Externalizable // externalizable for client code that may serialize this map
{
   private static final long serialVersionUID = 10929568968766L;
   private Entry<K, V>[] table;
   private int size;

   /**
    * For Externalizable
    */
   public ImmutableMapCopy()
   {
   }

   @SuppressWarnings("unchecked")
   public ImmutableMapCopy(Map<K, V> map)
   {
      size = map.size();
      table = new ImmutableEntry[size];
      int i = 0;
      for (Map.Entry<K, V> me : map.entrySet())
      {
         table[i++] = new ImmutableEntry<K, V>(me);
      }
   }

   @Override
   public int size()
   {
      return size;
   }

   @Override
   public final V remove(Object key)
   {
      throw new UnsupportedOperationException();
   }

   @Override
   public final void putAll(Map<? extends K, ? extends V> t)
   {
      throw new UnsupportedOperationException();
   }

   @Override
   public final void clear()
   {
      throw new UnsupportedOperationException();
   }

   public final Set<Entry<K, V>> entrySet()
   {
      return new ImmutableSetCopy<Map.Entry<K, V>>(table);
   }

   /**
    * Format:
    * - entry array size (int)
    * - entry key-value pairs (Object, Object)
    *
    * @param out stream to write to
    * @throws IOException
    */
   public void writeExternal(ObjectOutput out) throws IOException
   {
      out.writeInt(size);
      for (Entry e : table)
      {
         out.writeObject(e.getKey());
         out.writeObject(e.getValue());
      }
   }

   /**
    * See {@link #writeExternal(java.io.ObjectOutput)} for serialization format
    *
    * @param in stream
    * @throws IOException
    * @throws ClassNotFoundException
    */
   @SuppressWarnings("unchecked")
   public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException
   {
      size = in.readInt();
      table = new Entry[size];
      for (int i = 0; i < size; i++)
      {
         table[i] = new ImmutableEntry(in.readObject(), in.readObject());
      }
   }

   private static class ImmutableEntry<K, V> implements Map.Entry<K, V>
   {
      K k;
      V v;

      private ImmutableEntry(Map.Entry<K, V> entry)
      {
         k = entry.getKey();
         v = entry.getValue();
      }

      private ImmutableEntry(K k, V v)
      {
         this.k = k;
         this.v = v;
      }

      public K getKey()
      {
         return k;
      }

      public V getValue()
      {
         return v;
      }

      public V setValue(V value)
      {
         throw new UnsupportedOperationException();
      }
   }
}
