package org.jboss.cache.config.parsing;

import org.jboss.cache.Fqn;
import org.jboss.cache.RegionManagerImpl;
import org.jboss.cache.config.ConfigurationException;
import org.jboss.cache.config.EvictionConfig;
import org.jboss.cache.config.EvictionRegionConfig;
import org.jboss.cache.config.MissingPolicyException;
import org.jboss.cache.config.parsing.element.EvictionElementParser;
import org.jboss.cache.eviction.LRUConfiguration;
import org.jboss.cache.eviction.MRUConfiguration;
import org.testng.annotations.Test;
import org.w3c.dom.Element;

/**
 * Tester class for {@link org.jboss.cache.config.parsing.element.EvictionElementParser}.
 *
 * @author Mircea.Markus@jboss.com
 * @since 3.0
 */
@Test(groups = "unit")
public class EvictionElementParserTest
{
   EvictionElementParser parser = new EvictionElementParser();

   /**
    * Test that when nodes are not specified the config defaults to certain values.
    */
   public void testDefaults()
   {
      String xml =
            "   <eviction wakeUpInterval=\"5\">\n" +
                  "      <defaults policyClass=\"org.jboss.cache.eviction.MRUPolicy\"/>\n" +
                  "      <region name=\"/org/jboss/xyz\" policyClass=\"org.jboss.cache.eviction.MRUPolicy\" eventQueueSize=\"21\">\n" +
                  "         <attribute name=\"maxNodes\">2103</attribute>\n" +
                  "         <attribute name=\"minTimeToLive\">22</attribute>\n" +
                  "      </region>\n" +
                  "   </eviction>";
      EvictionConfig evictionConfig = getEvictionConfig(xml);
      assert evictionConfig.getDefaultEventQueueSize() == EvictionConfig.EVENT_QUEUE_SIZE_DEFAULT;
   }

   /**
    * test an happy flow.
    */
   public void testNormalConfig()
   {
      String xml =
            "   <eviction wakeUpInterval=\"5\" defaultPolicyClass=\"org.jboss.cache.eviction.MRUPolicy\" defaultEventQueueSize=\"123456\">\n" +
                  "      <default>\n" +
                  "         <attribute name=\"maxNodes\">6</attribute>\n" +
                  "         <attribute name=\"minTimeToLive\">7</attribute>\n" +
                  "      </default>\n" +
                  "      <region name=\"/org/jboss/data\">\n" +
                  "         <attribute name=\"minTimeToLive\">1002</attribute>\n" +
                  "         <attribute name=\"maxNodes\">2021</attribute>\n" +
                  "      </region>\n" +
                  "      <region name=\"/org/jboss/xyz\" policyClass=\"org.jboss.cache.eviction.LRUPolicy\" eventQueueSize=\"21\">\n" +
                  "         <attribute name=\"maxNodes\">2103</attribute>\n" +
                  "         <attribute name=\"timeToLive\">22</attribute>\n" +
                  "      </region>\n" +
                  "   </eviction>";
      EvictionConfig config = getEvictionConfig(xml);
      //tests the defaults
      assert config.getWakeupInterval() == 5;
      assert config.getDefaultEvictionPolicyClass().equals("org.jboss.cache.eviction.MRUPolicy");
      assert config.getDefaultEventQueueSize() == 123456;
      assert config.getEvictionRegionConfigs().size() == 3;

      //test first region config
      EvictionRegionConfig erConfig1 = config.getEvictionRegionConfigs().get(0);
      erConfig1.getRegionFqn().equals(Fqn.ROOT);
      MRUConfiguration defaultPolicyonfig = (MRUConfiguration) erConfig1.getEvictionPolicyConfig();
      assert defaultPolicyonfig.getMaxNodes() == 6;
      assert defaultPolicyonfig.getMinTimeToLive() == 7;

      //test second region config
      EvictionRegionConfig erConfig2 = config.getEvictionRegionConfigs().get(1);
      assert erConfig2.getEventQueueSize() == 123456;
      assert erConfig2.getRegionFqn().equals(Fqn.fromString("/org/jboss/data"));
      MRUConfiguration mruConfiguration = (MRUConfiguration) erConfig2.getEvictionPolicyConfig();
      assert mruConfiguration.getMinTimeToLive() == 1002;
      assert mruConfiguration.getMaxNodes() == 2021;

      //test 3rd region config
      EvictionRegionConfig erConfig3 = config.getEvictionRegionConfigs().get(2);
      assert erConfig3.getEventQueueSize() == 21;
      assert erConfig3.getRegionFqn().equals(Fqn.fromString("/org/jboss/xyz"));
      LRUConfiguration lruConfiguration = (LRUConfiguration) erConfig3.getEvictionPolicyConfig();
      assert lruConfiguration.getTimeToLive() == 22;
      assert lruConfiguration.getMaxNodes() == 2103;
   }

   public void testLruConfig()
   {
      String xml =
            "   <eviction wakeUpInterval=\"45000\" defaultPolicyClass=\"org.jboss.cache.eviction.LRUPolicy\" defaultEventQueueSize=\"4\">\n" +
                  "      <default>\n" +
                  "         <attribute name=\"maxNodes\">5000</attribute>\n" +
                  "         <attribute name=\"timeToLive\">1000000</attribute>\n" +
                  "         <attribute name=\"maxAge\">15000</attribute>\n" +
                  "      </default>\n" +
                  "      <region name=\"/fifo\">\n" +
                  "         <attribute name=\"maxNodes\">5000</attribute>\n" +
                  "         <attribute name=\"timeToLive\">1000000</attribute>\n" +
                  "      </region>\n" +
                  "      <region name=\"/mru\">\n" +
                  "         <attribute name=\"maxNodes\">10000</attribute>\n" +
                  "         <attribute name=\"timeToLive\">1000000</attribute>\n" +
                  "      </region>\n" +
                  "      <region name=\"/lfu\">\n" +
                  "         <attribute name=\"maxNodes\">5000</attribute>\n" +
                  "         <attribute name=\"timeToLive\">1000000</attribute>\n" +
                  "      </region>\n" +
                  "   </eviction>";
      EvictionConfig evConfig = getEvictionConfig(xml);
      EvictionRegionConfig evictionRegionConfig = evConfig.getEvictionRegionConfigs().get(0);
      assert evictionRegionConfig.getRegionName().equals(RegionManagerImpl.DEFAULT_REGION.toString());
      assert evictionRegionConfig.getRegionName().equals(RegionManagerImpl.DEFAULT_REGION.toString());
      assert ((LRUConfiguration) evictionRegionConfig.getEvictionPolicyConfig()).getTimeToLive() == 1000000;
   }

   /**
    * This is an mandatory parameter, if it is not present then an configuration exception is expected.
    */
   public void testMissingWakeUpInterval() throws Exception
   {
      String xml =
            "   <eviction defaultPolicyClass=\"org.jboss.cache.eviction.LRUPolicy\" defaultEventQueueSize=\"200000\">\n" +
                  "      <default>\n" +
                  "         <attribute name=\"maxNodes\">5000</attribute>\n" +
                  "         <attribute name=\"timeToLive\">1000</attribute>\n" +
                  "      </default>\n" +
                  "      <region name=\"/org/jboss/data\">\n" +
                  "         <attribute name=\"timeToLive\">1002</attribute>\n" +
                  "      </region>\n" +
                  "   </eviction>";
      try
      {
         getEvictionConfig(xml);
         assert false : "exception expected as wake up interval is not set";
      }
      catch (ConfigurationException e)
      {
         //expected
      }
   }

   /**
    * If no policy is configured at region basis, and also no policy is configured at region basis,
    * exception is expected.
    */
   public void testMissingPolicyOnRegion()
   {
      String xml =
            "   <eviction defaultEventQueueSize=\"200000\" wakeUpInterval=\"5000\">\n" +
                  "      <region name=\"/org/jboss/data\">\n" +
                  "         <attribute name=\"timeToLive\">1002</attribute>\n" +
                  "      </region>\n" +
                  "   </eviction>";
      try
      {
         getEvictionConfig(xml);
         assert false : "missing policy in both default and region, exception expected.";
      }
      catch (MissingPolicyException e)
      {
         //expected
      }
   }

   private EvictionConfig getEvictionConfig(String xml)
   {
      Element el;
      try
      {
         el = XmlConfigHelper.stringToElement(xml);
      }
      catch (Exception e)
      {
         throw new ConfigurationException(e);
      }
      return parser.parseEvictionElement(el);
   }

   public void testMissingDefaultEvictionClass() throws Exception
   {
      String xml =
            "   <eviction wakeUpInterval=\"5000\">\n" +
                  "      <default>\n" +
                  "         <attribute name=\"maxNodes\">5000</attribute>\n" +
                  "         <attribute name=\"timeToLive\">1000</attribute>\n" +
                  "      </default>\n" +
                  "      <region name=\"/org/jboss/data\" policyClass=\"org.jboss.cache.eviction.LFUPolicy\">\n" +
                  "         <attribute name=\"maxNodes\">5000</attribute>\n" +
                  "         <attribute name=\"minNodes\">1000</attribute>\n" +
                  "      </region>\n" +
                  "   </eviction>";
      try
      {
         getEvictionConfig(xml);
         assert false : " excewption expectecd as root does not have a eviction policy defined";
      }
      catch (MissingPolicyException e)
      {
         //expected
      }
   }
}
