package org.jboss.cache.eviction;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.RegionManagerImpl;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.EvictionConfig;
import org.jboss.cache.config.EvictionRegionConfig;
import org.jboss.cache.factories.UnitTestCacheConfigurationFactory;
import org.jboss.cache.interceptors.EvictionInterceptor;
import org.jboss.cache.transaction.DummyTransactionManagerLookup;
import org.jboss.cache.util.TestingUtil;
import static org.testng.AssertJUnit.assertFalse;
import static org.testng.AssertJUnit.assertTrue;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.transaction.TransactionManager;
import java.util.Iterator;
import java.util.List;

/**
 * Tests the eviction and the possible lack of locking nodes.
 * The configuration is with an aggressive eviction policy, 100 objects 2 seconds interval.
 * <p/>
 * It is possible that the number needs to be changed a little, depending on the machine speed.
 *
 * @author fhenning
 */
@Test(groups = {"functional"})
public class OptimisticEvictionTest
{

   //Maximum number of runs 2^20
   private static final int NUMBER_OF_RUNS = 1 << 20;
   //Initial number of nodes
   private static final int NUMBER_NODES = 256;

   private Fqn region = Fqn.fromElements("testingRegion");
   private TransactionManager txManager;
   private CacheSPI<Object, Object> cache;

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      txManager = new DummyTransactionManagerLookup().getTransactionManager();
      Configuration config = UnitTestCacheConfigurationFactory.createConfiguration(Configuration.CacheMode.LOCAL, true);
      config.setNodeLockingScheme(Configuration.NodeLockingScheme.OPTIMISTIC);
      config.setEvictionConfig(buildEvictionConfig());
      cache = (CacheSPI<Object, Object>) new DefaultCacheFactory<Object, Object>().createCache(config);
   }

   private EvictionConfig buildEvictionConfig() throws Exception
   {
      EvictionConfig result = new EvictionConfig("org.jboss.cache.eviction.LRUPolicy");
      result.setWakeupInterval(1000);

      LRUConfiguration lruConfiguration = new LRUConfiguration();
      lruConfiguration.setMaxNodes(10);
      lruConfiguration.setTimeToLive(0);
      lruConfiguration.setMaxAge(0);
      EvictionRegionConfig erConfig1 = new EvictionRegionConfig(RegionManagerImpl.DEFAULT_REGION, lruConfiguration);

      LRUConfiguration lruConfiguration2 = new LRUConfiguration();
      lruConfiguration2.setMaxNodes(10);
      lruConfiguration2.setTimeToLive(0);
      lruConfiguration2.setMaxAge(0);
      EvictionRegionConfig erConfig2 = new EvictionRegionConfig(Fqn.fromString("/testingRegion"), lruConfiguration2);

      LRUConfiguration lruConfiguration3 = new LRUConfiguration();
      lruConfiguration3.setMaxNodes(10);
      lruConfiguration3.setTimeToLive(1000);
      lruConfiguration3.setMaxAge(1000);
      EvictionRegionConfig erConfig3 = new EvictionRegionConfig(Fqn.fromString("/timeBased"), lruConfiguration3);

      result.getEvictionRegionConfigs().add(erConfig1);
      result.getEvictionRegionConfigs().add(erConfig2);
      result.getEvictionRegionConfigs().add(erConfig3);
      return result;
   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      if (cache != null)
      {
         // shut down any existing gtx2EntryMap
         try
         {
            if (cache.getTransactionManager().getTransaction() != null)
            {
               cache.getTransactionManager().rollback();
            }
         }
         catch (Exception e)
         {
            // ignore
         }
         cache.stop();
         cache = null;
      }
   }


   public void testEvictionError() throws Exception
   {
      //Initialize the cache via a map
      for (int i = 0; i < NUMBER_NODES; i++)
      {
         cache.put(Fqn.fromRelativeElements(region, i), i, i);
      }

      for (int i = 0; i < NUMBER_OF_RUNS; i++)
      {
         txManager.begin();
         cache.get(region, i % NUMBER_NODES);
         txManager.commit();
      }
   }


   public void testEvictionOccurence() throws Exception
   {
      cache.put("/timeBased/test", "key", "value");
      assertTrue(cache.exists("/timeBased/test"));

      // wait for it to be evicted.
      TestingUtil.sleepThread(3000);
      assertTrue(!cache.exists("/timeBased/test"));
   }

   public void testInterceptorChain() throws Exception
   {
      List interceptors = cache.getInterceptorChain();
      System.out.println(interceptors);
      Iterator i = interceptors.iterator();
      boolean found = false;

      while (i.hasNext())
      {
         Object o = i.next();
         if (o instanceof EvictionInterceptor)
         {
            found = true;
         }
      }

      assertTrue("Eviction interceptor should be in interceptor chain.", found);
   }

   public void testCompleteRemoval() throws Exception
   {
      String rootStr = "/timeBased/";

      // Add a parent, then a child. LRU will evict the parent,
      // then the child, leaving behind an empty parent
      Fqn parent = Fqn.fromString(rootStr + "parent");
      cache.put(parent, "key", "value");
      cache.put(Fqn.fromRelativeElements(parent, "child"), "key", "value");

      // Give eviction time to run a few times, then confirm parent
      // is completely gone
      TestingUtil.sleepThread(5500);
      assertFalse("Parent completely removed", cache.getRoot().hasChild(parent));
   }

}
